﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace TDP.TestAsyncAwait
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private void MakeUISafeCall(Form form, Action act)
        {
            if (form.InvokeRequired)
            {
                form.Invoke(act);
            }
            else
            {
                act.Invoke();
            }
        }

        private void BtnStartWorkNoTask_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Trace.WriteLine("BtnStartWorkNoTask_Click()->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

            for (int i = 0; i < 100000000; i++)
            {
                TxtCounter1.Text = i.ToString();
            }
        }

        private void BtnStartWorkTask_Click(object sender, EventArgs e)
        {
            Task T = new Task(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartWorkTask_Click->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                for (int i = 0; i < 10000000; i++)
                {
                    MakeUISafeCall(this, () => { TxtCounter1.Text = i.ToString(); });
                }
            });
            T.Start();
        }

        private void BtnStartTwoTasksBlocking_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Trace.WriteLine("BtnStartWorkTask_Click->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

            int T1Result;
            Task<int> T1 = new Task<int>(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartWorkTask_Click->Task 1->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                System.Threading.Thread.Sleep(5000);

                // Return a 'random' value
                return DateTime.Now.Second;
            });
            T1.Start();
            T1Result = T1.Result;   // We are waiting here and the application is blocked...

            int T2Result;
            Task<int> T2 = new Task<int>(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartWorkTask_Click->Task 2->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                System.Threading.Thread.Sleep(5000);

                return DateTime.Now.Second * T1Result;
            });
            T2.Start();
            T2Result = T2.Result;   // Again, we are waiting here and the application is blocked...

            TxtCounter1.Text = T2Result.ToString();
        }

        private async void BtnStartTwoTasksAsyncAwait_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Trace.WriteLine("BtnStartTwoTasksAsyncAwait_Click->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

            int T1Result;
            Task<int> T1 = new Task<int>(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartTwoTasksAsyncAwait_Click->Task 1->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                System.Threading.Thread.Sleep(5000);

                // Return a 'random' value
                return DateTime.Now.Second;
            });
            T1.Start();
            await T1;   // We are waiting for task to be completed, but the main thread is not blocked
            T1Result = T1.Result;

            int T2Result;
            Task<int> T2 = new Task<int>(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartTwoTasksAsyncAwait_Click->Task 2->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                System.Threading.Thread.Sleep(5000);

                return DateTime.Now.Second * T1Result;
            });
            T2.Start();
            await T2;   // Again we are waiting for task to be completed, but the main thread is not blocked
            T2Result = T2.Result;   

            TxtCounter1.Text = T2Result.ToString();
        }

        private Task<int> LongTaskWithExceptionAsync()
        {
            Task<int> T1 = new Task<int>(() =>
            {
                System.Diagnostics.Trace.WriteLine("BtnStartTaskWithException_Click->Task 1->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

                System.Threading.Thread.Sleep(5000);
                int a = 1, b = 1;

                if (a == b)
                    throw new ApplicationException("This is an exception!");

                // Return a 'random' value
                return DateTime.Now.Second;
            });
            T1.Start();
            return T1;
        }

        private async void BtnStartTaskWithException_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Trace.WriteLine("BtnStartTaskWithException_Click->Thread ID: " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString());

            try
            {
                int Val = await LongTaskWithExceptionAsync();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        private void BtnStartTaskWithAggrException_Click(object sender, EventArgs e)
        {
            try
            {
                Task<int> T = LongTaskWithExceptionAsync();
                T.Wait();
            }
            catch (AggregateException ex)
            {
                foreach (Exception inner_ex in ex.InnerExceptions)
                {
                    if (inner_ex is ApplicationException)
                    {
                        MessageBox.Show(inner_ex.ToString());
                    }
                    else
                    {
                        throw;
                    }
                }
            }
        }
    }
}
